import ctypes

from .api import library


def find_lib(lib):
    r"""
    Find the DLL for a given library.

    Accepts a string or loaded module

    >>> print(find_lib('kernel32').lower())
    c:\windows\system32\kernel32.dll
    """
    if isinstance(lib, str):
        lib = getattr(ctypes.windll, lib)

    size = 1024
    result = ctypes.create_unicode_buffer(size)
    library.GetModuleFileName(lib._handle, result, size)
    return result.value



from jaraco.windows.filesystem import trace_symlink_target

from optparse import OptionParser


def get_args():
    parser = OptionParser()
    options, args = parser.parse_args()
    try:
        options.filename = args.pop(0)
    except IndexError:
        parser.error('filename required')
    return options


def main():
    options = get_args()
    print(trace_symlink_target(options.filename))


if __name__ == '__main__':
    main()



from itertools import count

from six.moves import winreg  # type: ignore


def key_values(key):
    for index in count():
        try:
            yield winreg.EnumValue(key, index)
        except WindowsError:
            break


def key_subkeys(key):
    for index in count():
        try:
            yield winreg.EnumKey(key, index)
        except WindowsError:
            break



import ctypes.wintypes

GMEM_MOVEABLE = 0x2

GlobalAlloc = ctypes.windll.kernel32.GlobalAlloc
GlobalAlloc.argtypes = ctypes.wintypes.UINT, ctypes.c_size_t
GlobalAlloc.restype = ctypes.wintypes.HANDLE

GlobalLock = ctypes.windll.kernel32.GlobalLock
GlobalLock.argtypes = (ctypes.wintypes.HGLOBAL,)
GlobalLock.restype = ctypes.wintypes.LPVOID

GlobalUnlock = ctypes.windll.kernel32.GlobalUnlock
GlobalUnlock.argtypes = (ctypes.wintypes.HGLOBAL,)
GlobalUnlock.restype = ctypes.wintypes.BOOL

GlobalSize = ctypes.windll.kernel32.GlobalSize
GlobalSize.argtypes = (ctypes.wintypes.HGLOBAL,)
GlobalSize.restype = ctypes.c_size_t

CreateFileMapping = ctypes.windll.kernel32.CreateFileMappingW
CreateFileMapping.argtypes = [
    ctypes.wintypes.HANDLE,
    ctypes.c_void_p,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.LPWSTR,
]
CreateFileMapping.restype = ctypes.wintypes.HANDLE

MapViewOfFile = ctypes.windll.kernel32.MapViewOfFile
MapViewOfFile.restype = ctypes.wintypes.HANDLE

UnmapViewOfFile = ctypes.windll.kernel32.UnmapViewOfFile
UnmapViewOfFile.argtypes = (ctypes.wintypes.HANDLE,)

RtlMoveMemory = ctypes.windll.kernel32.RtlMoveMemory
RtlMoveMemory.argtypes = (ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t)

ctypes.windll.kernel32.LocalFree.argtypes = (ctypes.wintypes.HLOCAL,)



from ctypes import windll, POINTER
from ctypes.wintypes import LPWSTR, DWORD, LPVOID, HANDLE, BOOL

CreateEvent = windll.kernel32.CreateEventW
CreateEvent.argtypes = (LPVOID, BOOL, BOOL, LPWSTR)  # LPSECURITY_ATTRIBUTES
CreateEvent.restype = HANDLE

SetEvent = windll.kernel32.SetEvent
SetEvent.argtypes = (HANDLE,)
SetEvent.restype = BOOL

WaitForSingleObject = windll.kernel32.WaitForSingleObject
WaitForSingleObject.argtypes = HANDLE, DWORD
WaitForSingleObject.restype = DWORD

_WaitForMultipleObjects = windll.kernel32.WaitForMultipleObjects
_WaitForMultipleObjects.argtypes = DWORD, POINTER(HANDLE), BOOL, DWORD
_WaitForMultipleObjects.restype = DWORD


def WaitForMultipleObjects(handles, wait_all=False, timeout=0):
    n_handles = len(handles)
    handle_array = (HANDLE * n_handles)()
    for index, handle in enumerate(handles):
        handle_array[index] = handle
    return _WaitForMultipleObjects(n_handles, handle_array, wait_all, timeout)


WAIT_OBJECT_0 = 0
INFINITE = -1
WAIT_TIMEOUT = 0x102



import ctypes.wintypes

# from WinNT.h
READ_CONTROL = 0x00020000
STANDARD_RIGHTS_REQUIRED = 0x000F0000
STANDARD_RIGHTS_READ = READ_CONTROL
STANDARD_RIGHTS_WRITE = READ_CONTROL
STANDARD_RIGHTS_EXECUTE = READ_CONTROL
STANDARD_RIGHTS_ALL = 0x001F0000

# from NTSecAPI.h
POLICY_VIEW_LOCAL_INFORMATION = 0x00000001
POLICY_VIEW_AUDIT_INFORMATION = 0x00000002
POLICY_GET_PRIVATE_INFORMATION = 0x00000004
POLICY_TRUST_ADMIN = 0x00000008
POLICY_CREATE_ACCOUNT = 0x00000010
POLICY_CREATE_SECRET = 0x00000020
POLICY_CREATE_PRIVILEGE = 0x00000040
POLICY_SET_DEFAULT_QUOTA_LIMITS = 0x00000080
POLICY_SET_AUDIT_REQUIREMENTS = 0x00000100
POLICY_AUDIT_LOG_ADMIN = 0x00000200
POLICY_SERVER_ADMIN = 0x00000400
POLICY_LOOKUP_NAMES = 0x00000800
POLICY_NOTIFICATION = 0x00001000

POLICY_ALL_ACCESS = (
    STANDARD_RIGHTS_REQUIRED
    | POLICY_VIEW_LOCAL_INFORMATION
    | POLICY_VIEW_AUDIT_INFORMATION
    | POLICY_GET_PRIVATE_INFORMATION
    | POLICY_TRUST_ADMIN
    | POLICY_CREATE_ACCOUNT
    | POLICY_CREATE_SECRET
    | POLICY_CREATE_PRIVILEGE
    | POLICY_SET_DEFAULT_QUOTA_LIMITS
    | POLICY_SET_AUDIT_REQUIREMENTS
    | POLICY_AUDIT_LOG_ADMIN
    | POLICY_SERVER_ADMIN
    | POLICY_LOOKUP_NAMES
)


POLICY_READ = (
    STANDARD_RIGHTS_READ
    | POLICY_VIEW_AUDIT_INFORMATION
    | POLICY_GET_PRIVATE_INFORMATION
)

POLICY_WRITE = (
    STANDARD_RIGHTS_WRITE
    | POLICY_TRUST_ADMIN
    | POLICY_CREATE_ACCOUNT
    | POLICY_CREATE_SECRET
    | POLICY_CREATE_PRIVILEGE
    | POLICY_SET_DEFAULT_QUOTA_LIMITS
    | POLICY_SET_AUDIT_REQUIREMENTS
    | POLICY_AUDIT_LOG_ADMIN
    | POLICY_SERVER_ADMIN
)

POLICY_EXECUTE = (
    STANDARD_RIGHTS_EXECUTE | POLICY_VIEW_LOCAL_INFORMATION | POLICY_LOOKUP_NAMES
)


class TokenAccess:
    TOKEN_QUERY = 0x8


class TokenInformationClass:
    TokenUser = 1


class TOKEN_USER(ctypes.Structure):
    num = 1
    _fields_ = [('SID', ctypes.c_void_p), ('ATTRIBUTES', ctypes.wintypes.DWORD)]


class SECURITY_DESCRIPTOR(ctypes.Structure):
    """
    typedef struct _SECURITY_DESCRIPTOR
        {
        UCHAR Revision;
        UCHAR Sbz1;
        SECURITY_DESCRIPTOR_CONTROL Control;
        PSID Owner;
        PSID Group;
        PACL Sacl;
        PACL Dacl;
        }   SECURITY_DESCRIPTOR;
    """

    SECURITY_DESCRIPTOR_CONTROL = ctypes.wintypes.USHORT
    REVISION = 1

    _fields_ = [
        ('Revision', ctypes.c_ubyte),
        ('Sbz1', ctypes.c_ubyte),
        ('Control', SECURITY_DESCRIPTOR_CONTROL),
        ('Owner', ctypes.c_void_p),
        ('Group', ctypes.c_void_p),
        ('Sacl', ctypes.c_void_p),
        ('Dacl', ctypes.c_void_p),
    ]


class SECURITY_ATTRIBUTES(ctypes.Structure):
    """
    typedef struct _SECURITY_ATTRIBUTES {
        DWORD  nLength;
        LPVOID lpSecurityDescriptor;
        BOOL   bInheritHandle;
    } SECURITY_ATTRIBUTES;
    """

    _fields_ = [
        ('nLength', ctypes.wintypes.DWORD),
        ('lpSecurityDescriptor', ctypes.c_void_p),
        ('bInheritHandle', ctypes.wintypes.BOOL),
    ]

    def __init__(self, *args, **kwargs):
        super(SECURITY_ATTRIBUTES, self).__init__(*args, **kwargs)
        self.nLength = ctypes.sizeof(SECURITY_ATTRIBUTES)

    @property
    def descriptor(self):
        return self._descriptor

    @descriptor.setter
    def descriptor(self, value):
        self._descriptor = value
        self.lpSecurityDescriptor = ctypes.addressof(value)


ctypes.windll.advapi32.SetSecurityDescriptorOwner.argtypes = (
    ctypes.POINTER(SECURITY_DESCRIPTOR),
    ctypes.c_void_p,
    ctypes.wintypes.BOOL,
)



#!/usr/bin/env python

# $Id$

"""
jaraco.windows

A lightweight wrapper to provide a pythonic API to the Windows Platform.

This package attempts to provide interfaces similar or compatible
with Mark Hammond's pywin32 library, but avoids the use of extension
modules by utilizing ctypes.
"""



import subprocess
import itertools

from more_itertools import consume, always_iterable


def extract_environment(env_cmd, initial=None):
    """
    Take a command (either a single command or list of arguments)
    and return the environment created after running that command.
    Note that if the command must be a batch file or .cmd file, or the
    changes to the environment will not be captured.

    If initial is supplied, it is used as the initial environment passed
    to the child process.
    """
    # construct the command that will alter the environment
    env_cmd = subprocess.list2cmdline(always_iterable(env_cmd))
    # create a tag so we can tell in the output when the proc is done
    tag = 'Done running command'
    # construct a cmd.exe command to do accomplish this
    cmd = 'cmd.exe /s /c "{env_cmd} && echo "{tag}" && set"'.format(**vars())
    # launch the process
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=initial)
    # parse the output sent to stdout
    lines = proc.stdout
    # make sure the lines are strings

    def make_str(s):
        return s.decode()

    lines = map(make_str, lines)
    # consume whatever output occurs until the tag is reached
    consume(itertools.takewhile(lambda l: tag not in l, lines))
    # construct a dictionary of the pairs
    result = dict(line.rstrip().split('=', 1) for line in lines)
    # let the process finish
    proc.communicate()
    return result



import ctypes.wintypes

from jaraco.windows.error import handle_nonzero_success
from .api import security


def GetTokenInformation(token, information_class):
    """
    Given a token, get the token information for it.
    """
    data_size = ctypes.wintypes.DWORD()
    ctypes.windll.advapi32.GetTokenInformation(
        token, information_class.num, 0, 0, ctypes.byref(data_size)
    )
    data = ctypes.create_string_buffer(data_size.value)
    handle_nonzero_success(
        ctypes.windll.advapi32.GetTokenInformation(
            token,
            information_class.num,
            ctypes.byref(data),
            ctypes.sizeof(data),
            ctypes.byref(data_size),
        )
    )
    return ctypes.cast(data, ctypes.POINTER(security.TOKEN_USER)).contents


def OpenProcessToken(proc_handle, access):
    result = ctypes.wintypes.HANDLE()
    proc_handle = ctypes.wintypes.HANDLE(proc_handle)
    handle_nonzero_success(
        ctypes.windll.advapi32.OpenProcessToken(
            proc_handle, access, ctypes.byref(result)
        )
    )
    return result


def get_current_user():
    """
    Return a TOKEN_USER for the owner of this process.
    """
    process = OpenProcessToken(
        ctypes.windll.kernel32.GetCurrentProcess(), security.TokenAccess.TOKEN_QUERY
    )
    return GetTokenInformation(process, security.TOKEN_USER)


def get_security_attributes_for_user(user=None):
    """
    Return a SECURITY_ATTRIBUTES structure with the SID set to the
    specified user (uses current user if none is specified).
    """
    if user is None:
        user = get_current_user()

    assert isinstance(user, security.TOKEN_USER), "user must be TOKEN_USER instance"

    SD = security.SECURITY_DESCRIPTOR()
    SA = security.SECURITY_ATTRIBUTES()
    # by attaching the actual security descriptor, it will be garbage-
    # collected with the security attributes
    SA.descriptor = SD
    SA.bInheritHandle = 1

    ctypes.windll.advapi32.InitializeSecurityDescriptor(
        ctypes.byref(SD), security.SECURITY_DESCRIPTOR.REVISION
    )
    ctypes.windll.advapi32.SetSecurityDescriptorOwner(ctypes.byref(SD), user.SID, 0)
    return SA



import struct
import ctypes.wintypes
from ctypes.wintypes import DWORD, WCHAR, BYTE, BOOL


# from mprapi.h
MAX_INTERFACE_NAME_LEN = 2 ** 8

# from iprtrmib.h
MAXLEN_PHYSADDR = 2 ** 3
MAXLEN_IFDESCR = 2 ** 8

# from iptypes.h
MAX_ADAPTER_ADDRESS_LENGTH = 8
MAX_DHCPV6_DUID_LENGTH = 130


class MIB_IFROW(ctypes.Structure):
    _fields_ = (
        ('name', WCHAR * MAX_INTERFACE_NAME_LEN),
        ('index', DWORD),
        ('type', DWORD),
        ('MTU', DWORD),
        ('speed', DWORD),
        ('physical_address_length', DWORD),
        ('physical_address_raw', BYTE * MAXLEN_PHYSADDR),
        ('admin_status', DWORD),
        ('operational_status', DWORD),
        ('last_change', DWORD),
        ('octets_received', DWORD),
        ('unicast_packets_received', DWORD),
        ('non_unicast_packets_received', DWORD),
        ('incoming_discards', DWORD),
        ('incoming_errors', DWORD),
        ('incoming_unknown_protocols', DWORD),
        ('octets_sent', DWORD),
        ('unicast_packets_sent', DWORD),
        ('non_unicast_packets_sent', DWORD),
        ('outgoing_discards', DWORD),
        ('outgoing_errors', DWORD),
        ('outgoing_queue_length', DWORD),
        ('description_length', DWORD),
        ('description_raw', ctypes.c_char * MAXLEN_IFDESCR),
    )

    def _get_binary_property(self, name):
        val_prop = '{0}_raw'.format(name)
        val = getattr(self, val_prop)
        len_prop = '{0}_length'.format(name)
        length = getattr(self, len_prop)
        return str(memoryview(val))[:length]

    @property
    def physical_address(self):
        return self._get_binary_property('physical_address')

    @property
    def description(self):
        return self._get_binary_property('description')


class MIB_IFTABLE(ctypes.Structure):
    _fields_ = (
        ('num_entries', DWORD),  # dwNumEntries
        ('entries', MIB_IFROW * 0),  # table
    )


class MIB_IPADDRROW(ctypes.Structure):
    _fields_ = (
        ('address_num', DWORD),
        ('index', DWORD),
        ('mask', DWORD),
        ('broadcast_address', DWORD),
        ('reassembly_size', DWORD),
        ('unused', ctypes.c_ushort),
        ('type', ctypes.c_ushort),
    )

    @property
    def address(self):
        "The address in big-endian"
        _ = struct.pack('L', self.address_num)
        return struct.unpack('!L', _)[0]


class MIB_IPADDRTABLE(ctypes.Structure):
    _fields_ = (('num_entries', DWORD), ('entries', MIB_IPADDRROW * 0))


class SOCKADDR(ctypes.Structure):
    _fields_ = (('family', ctypes.c_ushort), ('data', ctypes.c_byte * 14))


LPSOCKADDR = ctypes.POINTER(SOCKADDR)


class SOCKET_ADDRESS(ctypes.Structure):
    _fields_ = [('address', LPSOCKADDR), ('length', ctypes.c_int)]


class _IP_ADAPTER_ADDRESSES_METRIC(ctypes.Structure):
    _fields_ = [('length', ctypes.c_ulong), ('interface_index', DWORD)]


class _IP_ADAPTER_ADDRESSES_U1(ctypes.Union):
    _fields_ = [
        ('alignment', ctypes.c_ulonglong),
        ('metric', _IP_ADAPTER_ADDRESSES_METRIC),
    ]


class IP_ADAPTER_ADDRESSES(ctypes.Structure):
    pass


LP_IP_ADAPTER_ADDRESSES = ctypes.POINTER(IP_ADAPTER_ADDRESSES)

# for now, just use void * for pointers to unused structures
PIP_ADAPTER_UNICAST_ADDRESS = ctypes.c_void_p
PIP_ADAPTER_ANYCAST_ADDRESS = ctypes.c_void_p
PIP_ADAPTER_MULTICAST_ADDRESS = ctypes.c_void_p
PIP_ADAPTER_DNS_SERVER_ADDRESS = ctypes.c_void_p
PIP_ADAPTER_PREFIX = ctypes.c_void_p
PIP_ADAPTER_WINS_SERVER_ADDRESS_LH = ctypes.c_void_p
PIP_ADAPTER_GATEWAY_ADDRESS_LH = ctypes.c_void_p
PIP_ADAPTER_DNS_SUFFIX = ctypes.c_void_p

IF_OPER_STATUS = ctypes.c_uint  # this is an enum, consider
# http://code.activestate.com/recipes/576415/
IF_LUID = ctypes.c_uint64

NET_IF_COMPARTMENT_ID = ctypes.c_uint32
GUID = ctypes.c_byte * 16
NET_IF_NETWORK_GUID = GUID
NET_IF_CONNECTION_TYPE = ctypes.c_uint  # enum
TUNNEL_TYPE = ctypes.c_uint  # enum

IP_ADAPTER_ADDRESSES._fields_ = [
    # ('u', _IP_ADAPTER_ADDRESSES_U1),
    ('length', ctypes.c_ulong),
    ('interface_index', DWORD),
    ('next', LP_IP_ADAPTER_ADDRESSES),
    ('adapter_name', ctypes.c_char_p),
    ('first_unicast_address', PIP_ADAPTER_UNICAST_ADDRESS),
    ('first_anycast_address', PIP_ADAPTER_ANYCAST_ADDRESS),
    ('first_multicast_address', PIP_ADAPTER_MULTICAST_ADDRESS),
    ('first_dns_server_address', PIP_ADAPTER_DNS_SERVER_ADDRESS),
    ('dns_suffix', ctypes.c_wchar_p),
    ('description', ctypes.c_wchar_p),
    ('friendly_name', ctypes.c_wchar_p),
    ('byte', BYTE * MAX_ADAPTER_ADDRESS_LENGTH),
    ('physical_address_length', DWORD),
    ('flags', DWORD),
    ('mtu', DWORD),
    ('interface_type', DWORD),
    ('oper_status', IF_OPER_STATUS),
    ('ipv6_interface_index', DWORD),
    ('zone_indices', DWORD),
    ('first_prefix', PIP_ADAPTER_PREFIX),
    ('transmit_link_speed', ctypes.c_uint64),
    ('receive_link_speed', ctypes.c_uint64),
    ('first_wins_server_address', PIP_ADAPTER_WINS_SERVER_ADDRESS_LH),
    ('first_gateway_address', PIP_ADAPTER_GATEWAY_ADDRESS_LH),
    ('ipv4_metric', ctypes.c_ulong),
    ('ipv6_metric', ctypes.c_ulong),
    ('luid', IF_LUID),
    ('dhcpv4_server', SOCKET_ADDRESS),
    ('compartment_id', NET_IF_COMPARTMENT_ID),
    ('network_guid', NET_IF_NETWORK_GUID),
    ('connection_type', NET_IF_CONNECTION_TYPE),
    ('tunnel_type', TUNNEL_TYPE),
    ('dhcpv6_server', SOCKET_ADDRESS),
    ('dhcpv6_client_duid', ctypes.c_byte * MAX_DHCPV6_DUID_LENGTH),
    ('dhcpv6_client_duid_length', ctypes.c_ulong),
    ('dhcpv6_iaid', ctypes.c_ulong),
    ('first_dns_suffix', PIP_ADAPTER_DNS_SUFFIX),
]

# define some parameters to the API Functions
GetIfTable = ctypes.windll.iphlpapi.GetIfTable
GetIfTable.argtypes = [
    ctypes.POINTER(MIB_IFTABLE),
    ctypes.POINTER(ctypes.c_ulong),
    BOOL,
]
GetIfTable.restype = DWORD

GetIpAddrTable = ctypes.windll.iphlpapi.GetIpAddrTable
GetIpAddrTable.argtypes = [
    ctypes.POINTER(MIB_IPADDRTABLE),
    ctypes.POINTER(ctypes.c_ulong),
    BOOL,
]
GetIpAddrTable.restype = DWORD

GetAdaptersAddresses = ctypes.windll.iphlpapi.GetAdaptersAddresses
GetAdaptersAddresses.argtypes = [
    ctypes.c_ulong,
    ctypes.c_ulong,
    ctypes.c_void_p,
    ctypes.POINTER(IP_ADAPTER_ADDRESSES),
    ctypes.POINTER(ctypes.c_ulong),
]
GetAdaptersAddresses.restype = ctypes.c_ulong



import ctypes.wintypes

SetEnvironmentVariable = ctypes.windll.kernel32.SetEnvironmentVariableW
SetEnvironmentVariable.restype = ctypes.wintypes.BOOL
SetEnvironmentVariable.argtypes = [ctypes.wintypes.LPCWSTR] * 2

GetEnvironmentVariable = ctypes.windll.kernel32.GetEnvironmentVariableW
GetEnvironmentVariable.restype = ctypes.wintypes.BOOL
GetEnvironmentVariable.argtypes = [
    ctypes.wintypes.LPCWSTR,
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.DWORD,
]



import ctypes.wintypes

try:
    from ctypes.wintypes import LPDWORD
except ImportError:
    LPDWORD = ctypes.POINTER(ctypes.wintypes.DWORD)  # type: ignore

GetUserName = ctypes.windll.advapi32.GetUserNameW
GetUserName.argtypes = ctypes.wintypes.LPWSTR, LPDWORD
GetUserName.restype = ctypes.wintypes.DWORD



import ctypes.wintypes

SystemParametersInfo = ctypes.windll.user32.SystemParametersInfoW
SystemParametersInfo.argtypes = (
    ctypes.wintypes.UINT,
    ctypes.wintypes.UINT,
    ctypes.c_void_p,
    ctypes.wintypes.UINT,
)

SPI_GETACTIVEWINDOWTRACKING = 0x1000
SPI_SETACTIVEWINDOWTRACKING = 0x1001
SPI_GETACTIVEWNDTRKTIMEOUT = 0x2002
SPI_SETACTIVEWNDTRKTIMEOUT = 0x2003



import ctypes.wintypes

# MPR - Multiple Provider Router
mpr = ctypes.windll.mpr

RESOURCETYPE_ANY = 0


class NETRESOURCE(ctypes.Structure):
    _fields_ = [
        ('scope', ctypes.wintypes.DWORD),
        ('type', ctypes.wintypes.DWORD),
        ('display_type', ctypes.wintypes.DWORD),
        ('usage', ctypes.wintypes.DWORD),
        ('local_name', ctypes.wintypes.LPWSTR),
        ('remote_name', ctypes.wintypes.LPWSTR),
        ('comment', ctypes.wintypes.LPWSTR),
        ('provider', ctypes.wintypes.LPWSTR),
    ]


LPNETRESOURCE = ctypes.POINTER(NETRESOURCE)

WNetAddConnection2 = mpr.WNetAddConnection2W
WNetAddConnection2.argtypes = (
    LPNETRESOURCE,
    ctypes.wintypes.LPCWSTR,
    ctypes.wintypes.LPCWSTR,
    ctypes.wintypes.DWORD,
)



import functools

from six.moves import map

import win32api
import win32evtlog
import win32evtlogutil

error = win32api.error  # The error the evtlog module raises.


class EventLog(object):
    def __init__(self, name="Application", machine_name=None):
        self.machine_name = machine_name
        self.name = name
        self.formatter = functools.partial(
            win32evtlogutil.FormatMessage, logType=self.name
        )

    def __enter__(self):
        if hasattr(self, 'handle'):
            raise ValueError("Overlapping attempts to use this log context")
        self.handle = win32evtlog.OpenEventLog(self.machine_name, self.name)
        return self

    def __exit__(self, *args):
        win32evtlog.CloseEventLog(self.handle)
        del self.handle

    _default_flags = (
        win32evtlog.EVENTLOG_BACKWARDS_READ | win32evtlog.EVENTLOG_SEQUENTIAL_READ
    )

    def get_records(self, flags=_default_flags):
        with self:
            while True:
                objects = win32evtlog.ReadEventLog(self.handle, flags, 0)
                if not objects:
                    break
                for item in objects:
                    yield item

    def __iter__(self):
        return self.get_records()

    def format_record(self, record):
        return self.formatter(record)

    def format_records(self, records=None):
        if records is None:
            records = self.get_records()
        return map(self.format_record, records)



"""
Python routines to interface with the Microsoft
Data Protection API (DPAPI).

>>> orig_data = b'Ipsum Lorem...'
>>> ciphertext = CryptProtectData(orig_data)
>>> descr, data = CryptUnprotectData(ciphertext)
>>> data == orig_data
True
"""

import ctypes
from ctypes import wintypes
from jaraco.windows.error import handle_nonzero_success


# for type declarations
__import__('jaraco.windows.api.memory')


class DATA_BLOB(ctypes.Structure):
    r"""
    A data blob structure for use with MS DPAPI functions.

    Initialize with string of characters
    >>> input = b'abc123\x00456'
    >>> blob = DATA_BLOB(input)
    >>> len(blob)
    10
    >>> blob.get_data() == input
    True
    """
    _fields_ = [('data_size', wintypes.DWORD), ('data', ctypes.c_void_p)]

    def __init__(self, data=None):
        super(DATA_BLOB, self).__init__()
        self.set_data(data)

    def set_data(self, data):
        "Use this method to set the data for this blob"
        if data is None:
            self.data_size = 0
            self.data = None
            return
        self.data_size = len(data)
        # create a string buffer so that null bytes aren't interpreted
        #  as the end of the string
        self.data = ctypes.cast(ctypes.create_string_buffer(data), ctypes.c_void_p)

    def get_data(self):
        "Get the data for this blob"
        array = ctypes.POINTER(ctypes.c_char * len(self))
        return ctypes.cast(self.data, array).contents.raw

    def __len__(self):
        return self.data_size

    def __str__(self):
        return self.get_data()

    def free(self):
        """
        "data out" blobs have locally-allocated memory.
        Call this method to free the memory allocated by CryptProtectData
        and CryptUnprotectData.
        """
        ctypes.windll.kernel32.LocalFree(self.data)


p_DATA_BLOB = ctypes.POINTER(DATA_BLOB)

_CryptProtectData = ctypes.windll.crypt32.CryptProtectData
_CryptProtectData.argtypes = [
    p_DATA_BLOB,  # data in
    wintypes.LPCWSTR,  # data description
    p_DATA_BLOB,  # optional entropy
    ctypes.c_void_p,  # reserved
    ctypes.c_void_p,  # POINTER(CRYPTPROTECT_PROMPTSTRUCT), # prompt struct
    wintypes.DWORD,  # flags
    p_DATA_BLOB,  # data out
]
_CryptProtectData.restype = wintypes.BOOL

_CryptUnprotectData = ctypes.windll.crypt32.CryptUnprotectData
_CryptUnprotectData.argtypes = [
    p_DATA_BLOB,  # data in
    ctypes.POINTER(wintypes.LPWSTR),  # data description
    p_DATA_BLOB,  # optional entropy
    ctypes.c_void_p,  # reserved
    ctypes.c_void_p,  # POINTER(CRYPTPROTECT_PROMPTSTRUCT), # prompt struct
    wintypes.DWORD,  # flags
    p_DATA_BLOB,  # data out
]
_CryptUnprotectData.restype = wintypes.BOOL

CRYPTPROTECT_UI_FORBIDDEN = 0x01


def CryptProtectData(
    data, description=None, optional_entropy=None, prompt_struct=None, flags=0
):
    """
    Encrypt data
    """
    data_in = DATA_BLOB(data)
    entropy = DATA_BLOB(optional_entropy) if optional_entropy else None
    data_out = DATA_BLOB()

    res = _CryptProtectData(
        data_in, description, entropy, None, prompt_struct, flags, data_out  # reserved
    )
    handle_nonzero_success(res)
    res = data_out.get_data()
    data_out.free()
    return res


def CryptUnprotectData(data, optional_entropy=None, prompt_struct=None, flags=0):
    """
    Returns a tuple of (description, data) where description is the
    the description that was passed to the CryptProtectData call and
    data is the decrypted result.
    """
    data_in = DATA_BLOB(data)
    entropy = DATA_BLOB(optional_entropy) if optional_entropy else None
    data_out = DATA_BLOB()
    ptr_description = wintypes.LPWSTR()
    res = _CryptUnprotectData(
        data_in,
        ctypes.byref(ptr_description),
        entropy,
        None,  # reserved
        prompt_struct,
        flags | CRYPTPROTECT_UI_FORBIDDEN,
        data_out,
    )
    handle_nonzero_success(res)
    description = ptr_description.value
    if ptr_description.value is not None:
        ctypes.windll.kernel32.LocalFree(ptr_description)
    res = data_out.get_data()
    data_out.free()
    return description, res



import os


def findpath(target, start=os.path.curdir):
    r"""
    Find a path from start to target where target is relative to start.

    >>> orig_wd = os.getcwd()
    >>> os.chdir('c:\\windows') # so we know what the working directory is

    >>> findpath('d:\\')
    'd:\\'

    >>> findpath('d:\\', 'c:\\windows')
    'd:\\'

    >>> findpath('\\bar', 'd:\\')
    'd:\\bar'

    >>> findpath('\\bar', 'd:\\foo') # fails with '\\bar'
    'd:\\bar'

    >>> findpath('bar', 'd:\\foo')
    'd:\\foo\\bar'

    >>> findpath('bar\\baz', 'd:\\foo')
    'd:\\foo\\bar\\baz'

    >>> findpath('\\baz', 'd:\\foo\\bar') # fails with '\\baz'
    'd:\\baz'

    Since we're on the C drive, findpath may be allowed to return
    relative paths for targets on the same drive. I use abspath to
    confirm that the ultimate target is what we expect.
    >>> os.path.abspath(findpath('\\bar'))
    'c:\\bar'

    >>> os.path.abspath(findpath('bar'))
    'c:\\windows\\bar'

    >>> findpath('..', 'd:\\foo\\bar')
    'd:\\foo'

    >>> findpath('..\\bar', 'd:\\foo')
    'd:\\bar'

    The parent of the root directory is the root directory.
    >>> findpath('..', 'd:\\')
    'd:\\'

    restore the original working directory
    >>> os.chdir(orig_wd)
    """
    return os.path.normpath(os.path.join(start, target))


def main():
    import sys

    if sys.argv[1:]:
        print(findpath(*sys.argv[1:]))
    else:
        import doctest

        doctest.testmod()


if __name__ == '__main__':
    main()



import ctypes.wintypes

CreateSymbolicLink = ctypes.windll.kernel32.CreateSymbolicLinkW
CreateSymbolicLink.argtypes = (
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.DWORD,
)
CreateSymbolicLink.restype = ctypes.wintypes.BOOLEAN

CreateHardLink = ctypes.windll.kernel32.CreateHardLinkW
CreateHardLink.argtypes = (
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.LPVOID,  # reserved for LPSECURITY_ATTRIBUTES
)
CreateHardLink.restype = ctypes.wintypes.BOOLEAN

GetFileAttributes = ctypes.windll.kernel32.GetFileAttributesW
GetFileAttributes.argtypes = (ctypes.wintypes.LPWSTR,)
GetFileAttributes.restype = ctypes.wintypes.DWORD

SetFileAttributes = ctypes.windll.kernel32.SetFileAttributesW
SetFileAttributes.argtypes = ctypes.wintypes.LPWSTR, ctypes.wintypes.DWORD
SetFileAttributes.restype = ctypes.wintypes.BOOL

MAX_PATH = 260

GetFinalPathNameByHandle = ctypes.windll.kernel32.GetFinalPathNameByHandleW
GetFinalPathNameByHandle.argtypes = (
    ctypes.wintypes.HANDLE,
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.DWORD,
)
GetFinalPathNameByHandle.restype = ctypes.wintypes.DWORD


class SECURITY_ATTRIBUTES(ctypes.Structure):
    _fields_ = (
        ('length', ctypes.wintypes.DWORD),
        ('p_security_descriptor', ctypes.wintypes.LPVOID),
        ('inherit_handle', ctypes.wintypes.BOOLEAN),
    )


LPSECURITY_ATTRIBUTES = ctypes.POINTER(SECURITY_ATTRIBUTES)

CreateFile = ctypes.windll.kernel32.CreateFileW
CreateFile.argtypes = (
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.DWORD,
    LPSECURITY_ATTRIBUTES,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.HANDLE,
)
CreateFile.restype = ctypes.wintypes.HANDLE
FILE_SHARE_READ = 1
FILE_SHARE_WRITE = 2
FILE_SHARE_DELETE = 4
FILE_FLAG_OPEN_REPARSE_POINT = 0x00200000
FILE_FLAG_BACKUP_SEMANTICS = 0x2000000
NULL = 0
OPEN_EXISTING = 3
FILE_ATTRIBUTE_READONLY = 0x1
FILE_ATTRIBUTE_HIDDEN = 0x2
FILE_ATTRIBUTE_DIRECTORY = 0x10
FILE_ATTRIBUTE_NORMAL = 0x80
FILE_ATTRIBUTE_REPARSE_POINT = 0x400
GENERIC_READ = 0x80000000
FILE_READ_ATTRIBUTES = 0x80
INVALID_HANDLE_VALUE = ctypes.wintypes.HANDLE(-1).value

INVALID_FILE_ATTRIBUTES = 0xFFFFFFFF

ERROR_NO_MORE_FILES = 0x12

VOLUME_NAME_DOS = 0

CloseHandle = ctypes.windll.kernel32.CloseHandle
CloseHandle.argtypes = (ctypes.wintypes.HANDLE,)
CloseHandle.restype = ctypes.wintypes.BOOLEAN


class WIN32_FIND_DATA(ctypes.wintypes.WIN32_FIND_DATAW):
    """
    _fields_ = [
        ("dwFileAttributes", DWORD),
        ("ftCreationTime", FILETIME),
        ("ftLastAccessTime", FILETIME),
        ("ftLastWriteTime", FILETIME),
        ("nFileSizeHigh", DWORD),
        ("nFileSizeLow", DWORD),
        ("dwReserved0", DWORD),
        ("dwReserved1", DWORD),
        ("cFileName", WCHAR * MAX_PATH),
        ("cAlternateFileName", WCHAR * 14)]
    ]
    """

    @property
    def file_attributes(self):
        return self.dwFileAttributes

    @property
    def creation_time(self):
        return self.ftCreationTime

    @property
    def last_access_time(self):
        return self.ftLastAccessTime

    @property
    def last_write_time(self):
        return self.ftLastWriteTime

    @property
    def file_size_words(self):
        return [self.nFileSizeHigh, self.nFileSizeLow]

    @property
    def reserved(self):
        return [self.dwReserved0, self.dwReserved1]

    @property
    def filename(self):
        return self.cFileName

    @property
    def alternate_filename(self):
        return self.cAlternateFileName

    @property
    def file_size(self):
        return self.nFileSizeHigh << 32 + self.nFileSizeLow


LPWIN32_FIND_DATA = ctypes.POINTER(ctypes.wintypes.WIN32_FIND_DATAW)

FindFirstFile = ctypes.windll.kernel32.FindFirstFileW
FindFirstFile.argtypes = (ctypes.wintypes.LPWSTR, LPWIN32_FIND_DATA)
FindFirstFile.restype = ctypes.wintypes.HANDLE
FindNextFile = ctypes.windll.kernel32.FindNextFileW
FindNextFile.argtypes = (ctypes.wintypes.HANDLE, LPWIN32_FIND_DATA)
FindNextFile.restype = ctypes.wintypes.BOOLEAN

ctypes.windll.kernel32.FindClose.argtypes = (ctypes.wintypes.HANDLE,)

SCS_32BIT_BINARY = 0  # A 32-bit Windows-based application
SCS_64BIT_BINARY = 6  # A 64-bit Windows-based application
SCS_DOS_BINARY = 1  # An MS-DOS-based application
SCS_OS216_BINARY = 5  # A 16-bit OS/2-based application
SCS_PIF_BINARY = 3  # A PIF file that executes an MS-DOS-based application
SCS_POSIX_BINARY = 4  # A POSIX-based application
SCS_WOW_BINARY = 2  # A 16-bit Windows-based application

_GetBinaryType = ctypes.windll.kernel32.GetBinaryTypeW
_GetBinaryType.argtypes = (
    ctypes.wintypes.LPWSTR,
    ctypes.POINTER(ctypes.wintypes.DWORD),
)
_GetBinaryType.restype = ctypes.wintypes.BOOL

FILEOP_FLAGS = ctypes.wintypes.WORD


class BY_HANDLE_FILE_INFORMATION(ctypes.Structure):
    _fields_ = [
        ('file_attributes', ctypes.wintypes.DWORD),
        ('creation_time', ctypes.wintypes.FILETIME),
        ('last_access_time', ctypes.wintypes.FILETIME),
        ('last_write_time', ctypes.wintypes.FILETIME),
        ('volume_serial_number', ctypes.wintypes.DWORD),
        ('file_size_high', ctypes.wintypes.DWORD),
        ('file_size_low', ctypes.wintypes.DWORD),
        ('number_of_links', ctypes.wintypes.DWORD),
        ('file_index_high', ctypes.wintypes.DWORD),
        ('file_index_low', ctypes.wintypes.DWORD),
    ]

    @property
    def file_size(self):
        return (self.file_size_high << 32) + self.file_size_low

    @property
    def file_index(self):
        return (self.file_index_high << 32) + self.file_index_low


GetFileInformationByHandle = ctypes.windll.kernel32.GetFileInformationByHandle
GetFileInformationByHandle.restype = ctypes.wintypes.BOOL
GetFileInformationByHandle.argtypes = (
    ctypes.wintypes.HANDLE,
    ctypes.POINTER(BY_HANDLE_FILE_INFORMATION),
)


class SHFILEOPSTRUCT(ctypes.Structure):
    _fields_ = [
        ('status_dialog', ctypes.wintypes.HWND),
        ('operation', ctypes.wintypes.UINT),
        ('from_', ctypes.wintypes.LPWSTR),
        ('to', ctypes.wintypes.LPWSTR),
        ('flags', FILEOP_FLAGS),
        ('operations_aborted', ctypes.wintypes.BOOL),
        ('name_mapping_handles', ctypes.wintypes.LPVOID),
        ('progress_title', ctypes.wintypes.LPWSTR),
    ]


_SHFileOperation = ctypes.windll.shell32.SHFileOperationW
_SHFileOperation.argtypes = [ctypes.POINTER(SHFILEOPSTRUCT)]
_SHFileOperation.restype = ctypes.c_int

FOF_ALLOWUNDO = 64
FOF_NOCONFIRMATION = 16
FO_DELETE = 3

ReplaceFile = ctypes.windll.kernel32.ReplaceFileW
ReplaceFile.restype = ctypes.wintypes.BOOL
ReplaceFile.argtypes = [
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.LPVOID,
    ctypes.wintypes.LPVOID,
]

REPLACEFILE_WRITE_THROUGH = 0x1
REPLACEFILE_IGNORE_MERGE_ERRORS = 0x2
REPLACEFILE_IGNORE_ACL_ERRORS = 0x4


class STAT_STRUCT(ctypes.Structure):
    _fields_ = [
        ('dev', ctypes.c_uint),
        ('ino', ctypes.c_ushort),
        ('mode', ctypes.c_ushort),
        ('nlink', ctypes.c_short),
        ('uid', ctypes.c_short),
        ('gid', ctypes.c_short),
        ('rdev', ctypes.c_uint),
        # the following 4 fields are ctypes.c_uint64 for _stat64
        ('size', ctypes.c_uint),
        ('atime', ctypes.c_uint),
        ('mtime', ctypes.c_uint),
        ('ctime', ctypes.c_uint),
    ]


_wstat = ctypes.windll.msvcrt._wstat
_wstat.argtypes = [ctypes.wintypes.LPWSTR, ctypes.POINTER(STAT_STRUCT)]
_wstat.restype = ctypes.c_int

FILE_NOTIFY_CHANGE_LAST_WRITE = 0x10

FindFirstChangeNotification = ctypes.windll.kernel32.FindFirstChangeNotificationW
FindFirstChangeNotification.argtypes = (
    ctypes.wintypes.LPWSTR,
    ctypes.wintypes.BOOL,
    ctypes.wintypes.DWORD,
)
FindFirstChangeNotification.restype = ctypes.wintypes.HANDLE

FindCloseChangeNotification = ctypes.windll.kernel32.FindCloseChangeNotification
FindCloseChangeNotification.argtypes = (ctypes.wintypes.HANDLE,)
FindCloseChangeNotification.restype = ctypes.wintypes.BOOL

FindNextChangeNotification = ctypes.windll.kernel32.FindNextChangeNotification
FindNextChangeNotification.argtypes = (ctypes.wintypes.HANDLE,)
FindNextChangeNotification.restype = ctypes.wintypes.BOOL

FILE_FLAG_OPEN_REPARSE_POINT = 0x00200000
IO_REPARSE_TAG_SYMLINK = 0xA000000C
FSCTL_GET_REPARSE_POINT = 0x900A8

LPDWORD = ctypes.POINTER(ctypes.wintypes.DWORD)
LPOVERLAPPED = ctypes.wintypes.LPVOID

DeviceIoControl = ctypes.windll.kernel32.DeviceIoControl
DeviceIoControl.argtypes = [
    ctypes.wintypes.HANDLE,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.LPVOID,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.LPVOID,
    ctypes.wintypes.DWORD,
    LPDWORD,
    LPOVERLAPPED,
]
DeviceIoControl.restype = ctypes.wintypes.BOOL


class REPARSE_DATA_BUFFER(ctypes.Structure):
    _fields_ = [
        ('tag', ctypes.c_ulong),
        ('data_length', ctypes.c_ushort),
        ('reserved', ctypes.c_ushort),
        ('substitute_name_offset', ctypes.c_ushort),
        ('substitute_name_length', ctypes.c_ushort),
        ('print_name_offset', ctypes.c_ushort),
        ('print_name_length', ctypes.c_ushort),
        ('flags', ctypes.c_ulong),
        ('path_buffer', ctypes.c_byte * 1),
    ]

    def get_print_name(self):
        wchar_size = ctypes.sizeof(ctypes.wintypes.WCHAR)
        arr_typ = ctypes.wintypes.WCHAR * (self.print_name_length // wchar_size)
        data = ctypes.byref(self.path_buffer, self.print_name_offset)
        return ctypes.cast(data, ctypes.POINTER(arr_typ)).contents.value

    def get_substitute_name(self):
        wchar_size = ctypes.sizeof(ctypes.wintypes.WCHAR)
        arr_typ = ctypes.wintypes.WCHAR * (self.substitute_name_length // wchar_size)
        data = ctypes.byref(self.path_buffer, self.substitute_name_offset)
        return ctypes.cast(data, ctypes.POINTER(arr_typ)).contents.value
